import datetime
import json

from pydeflate.pydeflate_config import PYDEFLATE_PATHS, logger


def _diff_from_today(date: datetime.datetime):
    """Compare to today"""

    today = datetime.datetime.today()

    return (today - date).days


def warn_updates():
    if not (PYDEFLATE_PATHS.data / "data_updates.json").exists():
        return

    with open(PYDEFLATE_PATHS.data / "data_updates.json") as file:
        updates = json.load(file)

    for source, date in updates.items():
        d = datetime.datetime.strptime(date, "%Y-%m-%d")
        if _diff_from_today(d) > 50:
            message = (
                f'\n\nThe underlying data for "{source}" has not been updated'
                f" in over {_diff_from_today(d)} days. \nIn order to use"
                " pydeflate with the most recent data, please run:\n"
                "`pydeflate.update_all_data()`"
            )
            logger.warning(message)


def update_update_date(source: str):
    """Update the most recent update date for data to today"""

    today = datetime.datetime.today().strftime("%Y-%m-%d")

    # Check to see if specified path contains an update file. Create one if not
    if not (PYDEFLATE_PATHS.data / "data_updates.json").exists():
        updates = {}
        with open(PYDEFLATE_PATHS.data / "data_updates.json", "w") as outfile:
            json.dump(updates, outfile)

    with open(PYDEFLATE_PATHS.data / "data_updates.json") as file:
        updates = json.load(file)

    updates[source] = today

    with open(PYDEFLATE_PATHS.data / "data_updates.json", "w") as outfile:
        json.dump(updates, outfile)


def update_all_data() -> None:
    """Run to update all underlying data."""

    from pydeflate.get_data.imf_data import IMF
    from pydeflate.get_data.oecd_data import OECD
    from pydeflate.get_data.wb_data import WorldBank

    data = {
        "IMF WEO Data": IMF().update,
        "OECD DAC data": OECD().update,
        "WorldBank data": WorldBank().update,
    }

    for source, func in data.items():
        func()
        logger.info(f"****Successfully updated {source}****\n")



import pandas as pd

from pydeflate.deflate.deflator import Deflator
from pydeflate.get_data.exchange_data import (
    ExchangeIMF,
    ExchangeOECD,
    ExchangeWorldBank,
)
from pydeflate.get_data.imf_data import IMF
from pydeflate.get_data.oecd_data import OECD
from pydeflate.get_data.wb_data import WorldBank
from pydeflate.pydeflate_config import logger
from pydeflate.utils import check_year_as_number, oecd_codes, to_iso3

ValidDeflatorSources: dict = {
    "oecd_dac": "dac_deflator",
    "dac": "dac_deflator",
    "wb": "world_bank",
    "world_bank": "world_bank",
    "imf": "imf",
}

ValidDeflatorMethods: dict = {
    "oecd_dac": ["dac_deflator"],
    "world_bank": ["gdp", "gdp_linked", "cpi"],
    "imf": ["gdp", "pcpi", "pcpie"],
}

ValidExchangeSources: dict = {
    "oecd_dac": "oecd_dac",
    "dac": "oecd_dac",
    "wb": "world_bank",
    "world_bank": "world_bank",
    "imf": "world_bank",
}

ValidExchangeMethods: dict = {
    "oecd_dac": ["implied"],
    "world_bank": ["yearly_average", "effective_exchange"],
    "imf": ["implied"],
}

ExchangeSources: dict = {
    "oecd_dac": {
        "implied": ExchangeOECD(method="implied"),
    },
    "world_bank": {
        "yearly_average": ExchangeWorldBank(method="yearly_average"),
        "effective_exchange": ExchangeWorldBank(method="effective_exchange"),
    },
    "imf": {
        "implied": ExchangeIMF(method="implied"),
    },
}

DeflatorSources: dict = {
    "oecd_dac": OECD,
    "world_bank": WorldBank,
    "imf": IMF,
}


def _validate_deflator_source(deflator_source) -> str:
    if deflator_source not in ValidDeflatorSources:
        raise ValueError(f'"{deflator_source=}" not valid')
    return deflator_source


def _validate_deflator_method(deflator_source, deflator_method) -> str:
    # Check if the method specified is part of the available deflators methods
    if (
        deflator_method is not None
        and deflator_method not in ValidDeflatorMethods[deflator_source]
    ):
        raise ValueError(f'"{deflator_method=}" not valid')

    # Default to the first method in the list of valid methods for the source
    if deflator_method is None:
        deflator_method = ValidDeflatorMethods[deflator_source][0]

    return deflator_method


def _validate_exchange_source(deflator_method, exchange_source) -> str:
    # Check if the source specified is part of the available exchange source
    if exchange_source is not None and exchange_source not in ValidExchangeSources:
        raise ValueError(f'"{exchange_source=}" not valid')

    # Default to the first source in the list of valid sources
    if exchange_source is None:
        exchange_source = ValidExchangeMethods[deflator_method][0]

    return exchange_source


def _validate_exchange_method(exchange_source, exchange_method) -> str:
    # Check if the method specified is part of the available exchange methods

    if (
        exchange_method is not None
        and exchange_method not in ValidExchangeMethods[exchange_source]
    ):
        raise ValueError(f'"{exchange_method=}" not valid')

    # Default to the first method in the list of valid methods for the source
    if exchange_method is None:
        exchange_method = ValidExchangeMethods[exchange_source][0]

    return exchange_method


def _create_id_col(df: pd.DataFrame, id_type: str, id_column: str) -> pd.DataFrame:
    """Create an id column.

    By default, if a country does not have a DAC deflator, the DAC, Total deflator is used.
    """

    if id_type == "DAC":
        df["id_"] = df[id_column].map(oecd_codes()).fillna("DAC")
    else:
        df = df.pipe(
            to_iso3, codes_col=id_column, target_col="id_", src_classification=id_type
        )

    return df


def deflate(
    df: pd.DataFrame,
    base_year: int,
    deflator_source: str = None,
    deflator_method: str | None = None,
    exchange_source: str | None = None,
    exchange_method: str | None = None,
    source_currency: str = "USA",
    target_currency: str = "USA",
    id_column: str = "iso_code",
    id_type: str = "ISO3",
    date_column: str = "date",
    source_column: str = "value",
    target_column: str = "value",
    to_current: bool = False,
    method: str | None = None,
    source: str | None = None,
    iso_column: str | None = None,
    source_col: str | None = None,
    target_col: str | None = None,
) -> pd.DataFrame:
    """Deflate amounts to a given currency - base year combination.

    Takes a DataFrame containing flows data and returns a DataFrame containing
    the deflated amounts. It can also convert from constant to current by specifying
    to_current as 'True'.

    Args:
        df: the DataFrame containing the flows column to be deflated. If multiple
            columns need to be deflated, the function needs to be called multiple times.
        base_year: If converting from current to constant, the target base year for the
            constant figures. If converting from constant, the base year of the data.
        deflator_source:{‘oecd_dac’, 'wb', 'imf'}
            The source of the data used to build the deflators. The value (and
            completeness) of the price deflators may change based on the source.
            Additionally, the OECD DAC data is only available for DAC donors.
        deflator_method:{'gdp', 'gdp_linked', 'cpi', 'pcpi', 'pcpie', None}
            The method used to calculate the price deflator:

            For World Bank (source == 'wb'), default is "gdp":

            •'gdp': using GDP deflators.

            •'gdp_linked': a GDP deflator series which has been linked to
            produce a consistent time series to counteract breaks in series
            over time due to changes in base years, sources or methodologies.

            •'cpi': using Consumer Price Index data

            For IMF (source == 'imf'), default is "gdp":

            •'pcpi': Consumer Price Index data

            •'pcpie': end-period Consumer Price Index (e.g. for December each year).

            For OECD DAC (source == 'oecd_dac'), default is "dac_deflator":
            •'dac_deflator': using the OECD DAC deflator

        exchange_source: The source of the exchange rates. If None, the exchange rates
        of the deflator_source will be used.

        exchange_method: The method used to calculate the exchange rates. If None, the
        default exchange method of the exchange_source will be used.

        source_currency: The iso3 code of the source currency. Note that deflators for EU countries
            are only in Euros from the year in which the Euro was adopted. To produce
            deflators only in euros, use 'emu'.

        target_currency :The iso3 code of the deflated amounts. It can be the same as the source
            currency. In cases where they are different, the exchange rate will be
            applied. To produce deflators only in euros, use 'emu'.

        id_column: The column containing the id codes (iso3 codes, for example) of the data's
        currency.

        id_type:The classification type for the id_column. By default, ISO3 but others are possible.
            Any options used in the Country Converter package are valid. For the OECD DAC
            classification, use 'DAC'.

        date_column:The column containing the date values. The column can contain years (int)
            or datetime objects.

        source_column:The column containing the data to be deflated.

        target_column: Column where the deflated data will be stored. It can be the same as the
            source column if a copy of the original isn't needed.

        to_current: If True, amounts will be treated as in constant prices and converted to
        current prices.

        iso_column:Provided for backwards compatibility. An alias for id_column
        source: Provided for backwards compatibility. An alias for deflator_source
        method: Provided for backwards compatibility. An alias for deflator_method
        source_col: Provided for backwards compatibility. An alias for source_column
        target_col: Provided for backwards compatibility. An alias for target_column

    Returns:
        A pandas DataFrame containing the deflated data. Years for which there
        are no deflators will be returned as null values.


    """

    # ------------------ Backwards compatibility ------------------

    # Backwards compatibility: if the iso_column parameter is used, reassign it
    if iso_column is not None:
        id_column = iso_column
        logger.warning("iso_column is deprecated. Use id_column instead.")

    # Backwards compatibility: if the source parameter is used, reassign it
    if deflator_source is None and source is not None:
        deflator_source = source
        logger.warning("source is deprecated. Use deflator_source instead.")

    # Backwards compatibility: if the method parameter is used, reassign it
    if method is not None:
        deflator_method = method
        logger.warning("method is deprecated. Use deflator_method instead.")

    # Backwards compatibility: if the source_col parameter is used, reassign it
    if source_col is not None:
        source_column = source_col
        logger.warning("source_col is deprecated. Use source_column instead.")

    # Backwards compatibility: if the target_col parameter is used, reassign it
    if target_col is not None:
        target_column = target_col
        logger.warning("target_col is deprecated. Use target_column instead.")

    # -------------------------- Validation -----------------------------
    deflator_source = _validate_deflator_source(deflator_source=deflator_source)

    deflator_method = _validate_deflator_method(
        deflator_source=deflator_source, deflator_method=deflator_method
    )

    exchange_source = _validate_exchange_source(
        deflator_method=exchange_method, exchange_source=exchange_source
    )

    exchange_method = _validate_exchange_method(
        exchange_source=exchange_source, exchange_method=exchange_method
    )

    # ----------------------------- Set up ---------------------------

    # copy the dataframe to avoid modifying the original
    df = df.copy(deep=True)

    # Keep track of original columns to return data in the same order.
    if target_column not in df.columns:
        cols = [*df.columns, target_column]
    else:
        cols = df.columns

    # check if date format matches
    df, year_as_number = check_year_as_number(df=df, date_column=date_column)

    # create id_column
    df = _create_id_col(df=df, id_column=id_column, id_type=id_type)

    # ----------------------------- Load deflator data ---------------------------
    deflator = (
        Deflator(
            base_year=base_year,
            exchange_obj=ExchangeSources[exchange_source][exchange_method],
            deflator_obj=DeflatorSources[deflator_source](),
            deflator_method=deflator_method,
            source_currency=source_currency,
            target_currency=target_currency,
            to_current=to_current,
        )
        .get_deflator()
        .rename(columns={"iso_code": "id_", "year": date_column})
    )

    # ----------------------------- Deflate ---------------------------
    # Merge the original data with the deflator DataFrame
    df = df.merge(
        deflator,
        on=["id_", date_column],
        how="left",
        suffixes=("", "_"),
    )

    # If the reverse option is used, multiply the data. Else divide.
    df[target_column] = df[source_column] / (df.deflator / 100)

    # Keep only the columns present in the original DataFrame (including order)
    df = df.filter(cols, axis=1)

    # If the year is passed as a number, convert it back to a number.
    if year_as_number:
        df[date_column] = df[date_column].dt.year

    return df



from abc import ABC, abstractmethod
from dataclasses import dataclass

import pandas as pd

from pydeflate.get_data.imf_data import IMF
from pydeflate.get_data.wb_data import END, START, update_world_bank_data
from pydeflate.pydeflate_config import PYDEFLATE_PATHS, logger
from pydeflate.utils import emu


def _exchange_ratio(source_xe: pd.DataFrame, target_xe: pd.DataFrame) -> pd.DataFrame:
    return (
        source_xe.merge(
            target_xe,
            how="left",
            on=["year", "iso_code"],
            suffixes=("_source", "_target"),
        )
        .assign(value=lambda d: d.value_source / d.value_target)
        .drop(["value_source", "value_target"], axis=1)
    )


def _calculate_deflator(xe: pd.DataFrame, base_year: int) -> pd.DataFrame:
    # get the data for the selected base year.
    xe_base = xe.query(f"year.dt.year == {base_year}")

    # If base year is not valid, raise error
    if len(xe_base) == 0:
        raise ValueError(f"No currency exchange data for {base_year=}")

    # Merge the exchange data and the base year data
    xe = xe.merge(xe_base, how="left", on=["iso_code"], suffixes=("", "_base"))

    # Calculate the deflator
    xe.value = round(100 * xe.value / xe.value_base, 6)

    return xe.filter(["iso_code", "year", "value"], axis=1)


@dataclass
class Exchange(ABC):
    """An abstract class to update, load and return exchange rate data.

    Attributes:
        method: the method to use to calculate the exchange rate.

    """

    method: str | None
    _data: pd.DataFrame | None = None

    @abstractmethod
    def update(self, **kwargs) -> None:
        """Update underlying data"""

    @abstractmethod
    def load_data(self, **kwargs) -> None:
        """Load required data to construct deflator"""

    @abstractmethod
    def usd_exchange_rate(self, direction: str = "lcu_to_uds") -> pd.DataFrame:
        """Get the exchange rate of a currency to USD (or vice versa)

        Args:
            direction: the direction of the exchange rate. Either "lcu_to_usd"
            or "usd_to_lcu".

        Returns:
            A dataframe with the exchange rate of the currency to USD (or vice versa).
        """

    def exchange_rate(self, currency_iso: str) -> pd.DataFrame:
        """Get an exchange rate for a given ISO

        Args:
            currency_iso: the iso_code of the currency to get the exchange rate for.

        Returns:
            A dataframe with the exchange rate for the specified currency.
        """

        # get usd exchange rates
        d_ = self.usd_exchange_rate()

        # filter the exchange rate of the chosen currency
        target_iso = d_.query(f"iso_code == '{currency_iso}'")

        # If currency is not valid, raise error
        if len(target_iso) == 0:
            raise ValueError(f"No currency exchange data for {currency_iso=}")

        # merge the two datasets and calculate new exchange rate
        return (
            pd.merge(d_, target_iso, how="left", on=["year"], suffixes=("", "_xe"))
            .assign(value=lambda d: d.value / d.value_xe)
            .drop(columns=["value_xe", "iso_code_xe"])
        )

    def exchange_deflator(
        self, source_iso: str, target_iso: str, base_year: int
    ) -> pd.DataFrame:
        """Get an exchange rate deflator for a given ISO

        Args:
            source_iso: the iso_code in which the "original" data is based.
                The default should be "LCU".
            target_iso: the iso_code of the currency to get the exchange rate for.
            base_year: the base year to calculate the deflator.

        Returns:
            A dataframe with the exchange rate deflator for the specified currency.

        """
        # Get exchange data based on the source currency. If LCU set to None
        source_xe = (
            self.exchange_rate(currency_iso=source_iso)
            if source_iso != "LCU"
            else self.exchange_rate(currency_iso="USA").assign(value=1)
        )

        # Get exchange data based on the target currency. If LCU set to None
        target_xe = (
            self.exchange_rate(currency_iso=target_iso)
            if target_iso != "LCU"
            else self.exchange_rate(currency_iso="USA").assign(value=1)
        )

        # calculate conversion ratio
        exchange_ratio = _exchange_ratio(source_xe=source_xe, target_xe=target_xe)

        # get the data for the selected base year.
        if target_iso != "LCU":
            xe = self.exchange_rate(currency_iso=target_iso)
        elif source_iso != "LCU":
            xe = self.exchange_rate(currency_iso=source_iso)
        else:
            xe = self.exchange_rate(currency_iso="USA").assign(value=1)

        # Calculate the deflator
        xe = _calculate_deflator(xe=xe, base_year=base_year)

        # Merge the conversion ratio
        xe = xe.merge(
            exchange_ratio,
            how="left",
            on=["year", "iso_code"],
            suffixes=("_xe", "_xe_ratio"),
        )

        # Convert deflators to currency of interest
        xe = xe.assign(value=lambda d: d.value_xe * d.value_xe_ratio)

        return xe.filter(["year", "iso_code", "value"], axis=1)


@dataclass
class ExchangeOECD(Exchange):
    """The OECD exchange rate data

    Attributes:
        method: the method to use to calculate the exchange rate. For this source,
        the only valid method is “implied”.
    """

    method: str = "implied"
    _load_try_count: int = 0

    @staticmethod
    def update(**kwargs) -> None:
        """Update the DAC1 data, which is the source for the OECD exchange rates."""
        from pydeflate.get_data import oecd_data

        oecd_data.update_dac1()

    def load_data(self, **kwargs) -> None:
        """Load the DAC1 data, which is the source for the OECD exchange rates."""
        # avoid infinite recursion
        if self._load_try_count < 1:
            try:
                self._data = pd.read_feather(PYDEFLATE_PATHS.data / "dac1.feather")
            except FileNotFoundError:
                logger.info("OECD Data not found, downloading...")
                self._load_try_count = +1
                self.update()
                self.load_data()
        else:
            raise FileNotFoundError("Could not load OECD data")

    def usd_exchange_rate(self, direction: str = "lcu_usd") -> pd.DataFrame:
        """Get the exchange rate of a currency to USD (or vice versa)

        Args:
            direction: the direction of the exchange rate. Either "lcu_to_usd"
            or "usd_to_lcu".

        Returns:
            A dataframe with the exchange rate of the currency to USD (or vice versa).
        """

        # If data is not loaded, load it
        if self._data is None:
            self.load_data()

        # Filter the data to only include the required columns, rename the value column.
        d_ = self._data[["iso_code", "year", "exchange"]].rename(
            columns={"exchange": "value"}
        )

        # if the direction is lcu to usd, return the data
        if direction == "lcu_usd":
            return d_
        # if the direction is usd to lcu, invert the exchange rate.
        elif direction == "usd_lcu":
            d_.value = 1 / d_.value
            return d_
        # if the direction is not valid, raise error
        else:
            raise ValueError("Direction must be 'lcu_usd' or 'usd_lcu'")


@dataclass
class ExchangeWorldBank(Exchange):
    """The World Bank exchange rate data

    Attributes:
        method: the method to use to calculate the exchange rate. For this source,
        the valid methods are "yearly_average" and "effective_exchange".
    """

    method: str = "yearly_average"
    _load_try_count: int = 0

    def __post_init__(self):
        """Check that the method is valid"""
        if self.method not in ["yearly_average", "effective_exchange"]:
            raise ValueError("Method must be 'yearly_average' or 'effective_exchange'")

    def update(self) -> None:
        """Update the World Bank data"""
        update_world_bank_data()

    def load_data(self, **kwargs) -> None:
        """Load the World Bank data"""
        # method mapping
        method_map = {
            "yearly_average": "PA.NUS.FCRF",
            "effective_exchange": "PX.REX.REER",
        }

        if self._load_try_count < 1:
            # Try to load the data from disk
            try:
                self._data = (
                    pd.read_csv(
                        PYDEFLATE_PATHS.data
                        / f"{method_map[self.method]}_{START}-{END}_.csv",
                        parse_dates=["date"],
                    )
                    .rename(columns={"date": "year"})
                    .drop(columns=["indicator_code"])
                )

            # If the data is not found, download it. Reload the data to the object.
            except FileNotFoundError:
                logger.info("World Bank data not found, downloading...")
                self._load_try_count = +1
                self.update()
                self.load_data()
        else:
            raise FileNotFoundError("Could not load World Bank data")

    def usd_exchange_rate(self, direction: str = "lcu_usd") -> pd.DataFrame:
        """Get the exchange rate of a currency to USD (or vice versa)

        Args:
            direction: the direction of the exchange rate. Either "lcu_to_usd"
            or "usd_to_lcu".

        Returns:
            A dataframe with the exchange rate of the currency to USD (or vice versa).

        """

        if direction not in ["lcu_usd", "usd_lcu"]:
            raise ValueError("Direction must be 'lcu_usd' or 'usd_lcu'")

        # If data is not loaded, load it
        if self._data is None:
            self.load_data()

        # Load the data into a temporary variable
        df = self._data.copy(deep=True)

        # Find the "Euro" data. This is done given that some countries are missing
        # exchange rates, but they are part of the Euro area.
        eur = df.loc[df.iso_code == "EMU"].dropna().set_index("year")["value"].to_dict()

        # Euro area countries without exchange rates
        eur_mask = (df.iso_code.isin(emu())) & (df.value.isna())

        # Assign EURO exchange rate to euro are countries from year euro adopted
        df.loc[eur_mask, "value"] = df.year.map(eur)

        # If the direction is lcu to usd, return the data

        if direction == "lcu_usd":
            return df
        # if the direction is usd to lcu, invert the exchange rate.

        df.value = 1 / df.value
        return df


@dataclass
class ExchangeIMF(Exchange):
    """The IMF exchange rate data

    Attributes:
        method: the method to use to calculate the exchange rate. For this source,
        the only valid method is “implied”.

    """

    method: str = "implied"
    _imf: IMF | None = None

    def __post_init__(self):
        """Load an IMF object to manage getting the data"""
        if self._imf is None:
            self._imf = IMF()

    def update(self, **kwargs) -> None:
        """Update the IMF data"""
        self._imf.update()

    def load_data(self, **kwargs) -> None:
        """Load the IMF data"""

        # if no direction is passed, default to lcu to usd
        direction = "lcu_usd" if "direction" not in kwargs else kwargs["direction"]
        self._data = self._imf.implied_exchange(direction=direction)

    def usd_exchange_rate(self, direction: str = "lcu_usd") -> pd.DataFrame:
        """Get the exchange rate of a currency to USD (or vice versa)

        Args:
            direction: the direction of the exchange rate. Either "lcu_to_usd"
            or "usd_to_lcu".

        Returns:
            A dataframe with the exchange rate of the currency to USD (or vice versa).

        """

        if direction not in ["lcu_usd", "usd_lcu"]:
            raise ValueError("Direction must be 'lcu_usd' or 'usd_lcu'")

        # load the data
        self.load_data(direction=direction)

        # return the data
        return self._data



from pathlib import Path
import logging


class PYDEFLATE_PATHS:
    """Class to store the paths to the data and output folders."""

    package = Path(__file__).resolve().parent.parent
    data = package / "pydeflate" / ".pydeflate_data"
    settings = package / "pydeflate" / "settings"
    test_data = package / "tests" / "test_files"


# Create a root logger
logger = logging.getLogger(__name__)

# Create two handlers (terminal and file)
shell_handler = logging.StreamHandler()

# Set levels for the logger, shell and file
logger.setLevel(logging.DEBUG)
shell_handler.setLevel(logging.DEBUG)

# Format the outputs
fmt_shell = "%(levelname)s [%(filename)s: %(funcName)s:] %(message)s"

# Create formatters
shell_formatter = logging.Formatter(fmt_shell)

# Add formatters to handlers
shell_handler.setFormatter(shell_formatter)

# Add handlers to the logger
logger.addHandler(shell_handler)



from __future__ import annotations

from abc import ABC, abstractmethod
from dataclasses import dataclass

import pandas as pd


@dataclass
class Data(ABC):
    _data: pd.DataFrame | None = None
    _available_methods: dict | None = None
    """Abstract class defining the basic structure and functionality of Data classes.
    
    Data classes store the price data from different sources.
    """

    @abstractmethod
    def update(self, **kwargs) -> None:
        """Update underlying data"""

    @abstractmethod
    def load_data(self, **kwargs) -> None:
        """Load required data to construct deflator"""

    def available_methods(self) -> dict:
        """Return a dictionary of available methods with their functions"""
        return self._available_methods

    def get_method(self, method: str) -> pd.DataFrame:
        """Return the data for a given method"""
        if self._data is None:
            self.load_data()

        if method not in self._data.indicator.unique():
            raise ValueError(f"Invalid method {method=}")

        return (
            self._data.loc[lambda d: d.indicator == method]
            .filter(["iso_code", "year", "value"], axis=1)
            .sort_values(["iso_code", "year"])
            .reset_index(drop=True)
        )

    def get_deflator(self, base_year: int, method: str | None = None) -> pd.DataFrame:
        if method is None:
            raise ValueError("method must be specified")

        if method not in self.available_methods():
            raise ValueError(f"method must be one of {self.available_methods().keys()}")

        if self._data is None:
            self.load_data()

        d_ = self.get_method(self._available_methods[method])

        # get the data for the selected base year.
        d_base = d_.query(f"year.dt.year == {base_year}")

        # If base year is not valid, raise error
        if len(d_base) == 0:
            raise ValueError(f"No currency exchange data for {base_year=}")

        # Merge the exchange data and the base year data
        d_ = d_.merge(d_base, how="left", on=["iso_code"], suffixes=("", "_base"))

        # Calculate the deflator
        d_.value = round(100 * d_.value / d_.value_base, 6)

        return d_.filter(["iso_code", "year", "value"], axis=1)
